#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# (C) Copyright 2009, Markus Korn <thekorn@gmx.de>
#
# ##################################################################
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##################################################################

import os
import logging
import errno

from fusepy import FUSE, Operations # http://code.google.com/p/fusepy/
from zeitgeistfs.fselements import Root, Starter, FileProxy, ContentFile

_filesystem_in_logger = logging.getLogger("zeitgeistfs.FS.in")
_filesystem_out_logger = logging.getLogger("zeitgeistfs.FS.out")

class LoggingMixIn(object):
    def __call__(self, op, path, *args):
        _filesystem_in_logger.debug(" ".join(map(repr, (op, path, args))))
        try:
            ret = getattr(self, op)(path, *args)
            _filesystem_out_logger.debug(" ".join(map(repr, (op, ret))))
            return ret
        except OSError, e:
            _filesystem_out_logger.error(str(e))
            raise
        except:
            _filesystem_out_logger.exception("Unexpected Error:")
            raise

class ZeitgeistFS(LoggingMixIn, Operations):
    
    def __init__(self, mountpoint=""):
        self.__mountpoint = mountpoint
        self.__filesystem = Root(mountpoint=mountpoint, desktop=False)
        self.__fd = 10000 # don't want to interfere with systems filenos
        self.__buffers = {}
    
    def readdir(self, path, fh):
        return self.__filesystem[path].get_items()
        
    def getattr(self, path, fh=None):
        return self.__filesystem[path].getattr(fh)
        
    def readlink(self, path):
        target = self.__filesystem[path]
        target = os.path.join(self.__mountpoint, target.target.lstrip("/"))
        return target
        
    def open(self, path, flags):
        fileobj = self.__filesystem[path]
        if isinstance(fileobj, (ContentFile, Starter)):
            self.__fd += 1
            fd = self.__fd
        elif isinstance(fileobj, FileProxy):
            fd = fileobj.open(flags)
        else:
            raise OSError(errno.EACCES, "")
        self.__buffers[fd] = fileobj
        return fd
        
    def read(self, path, size, offset, fh):
        if not fh in self.__buffers:
            raise OSError(errno.EACCES, "")
        return self.__buffers[fh].read(size, offset)
        
    def flush(self, path, fh):
        self.__buffers[fh].flush()
        return 0
        
    def release(self, path, fh):
        self.__buffers[fh].release()
        del self.__buffers[fh]
        return 0
        
    def write(self, path, data, offset, fh):
        if not fh in self.__buffers:
            raise OSError(errno.EACCES, "")
        fileobj = self.__buffers[fh]
        try:
            return fileobj.write(data, offset)
        except ValueError:
            raise OSError(errno.EINVAL, "")
        except NotImplementedError:
            raise OSError(errno.EACCES, "")
        
    def truncate(self, path, length, fh=None):
        if fh is not None:
            if not fh in self.__buffers:
                raise OSError(errno.EACCES, "")
            fileobj = self.__buffers[fh]
        else:
            fileobj = self.__filesystem[path]
        try:
            return fileobj.truncate(length)
        except NotImplementedError:
            raise OSError(errno.EACCES, "")
