#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# (C) Copyright 2009, Markus Korn <thekorn@gmx.de>
#
# ##################################################################
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##################################################################

import stat
import time
import os.path
import calendar
import errno
import logging
import os
import string
import random
import urlparse
import textwrap
import json

from datetime import date

from zeitgeistfs import (JOURNAL_IDENTIFIER, TODAY_IDENTIFIER,
    TODAY_NAME, ROOT_JOURNAL, ROOT_TODAY)
from zeitgeistfs.desktopfile import StarterLink, WWWLink

_logger = logging.getLogger("zeitgeistfs.fselements")

_IFACE = None
def get_dbus_interface():
    global _IFACE
    if _IFACE is None:
        from zeitgeist.dbusutils import DBusInterface
        _IFACE = DBusInterface()
    return _IFACE
    
def drop_root(uri, length):
    return uri[length:].lstrip("/")

_PRINTABLES = string.letters + string.digits
def get_random_name(length=10, prefix="", suffix=""):
    return str(prefix) + "".join(random.sample(_PRINTABLES, length)) + str(suffix)


class _urldict(dict):
    
    def __init__(self):
        super(_urldict, self).__init__()
        self.__order = list()
        
    def sorted_iter(self):
        for key in self.__order:
            try:
                yield key, self[key]
            except KeyError:
                pass
    
    def get_or_list(self, key):
        result = self.setdefault(key, list())
        if result == list():
            self.__order.append(key)
        return result
        

class Item(object):
    
    def __init__(self, name, mountpoint=""):
        self.__name = name
        self.__mountpoint = mountpoint
        
    @property
    def name(self):
        return self.__name
        
    @property
    def mountpoint(self):
        return self.__mountpoint
        
    def getattr(self, fh=None):
        raise NotImplementedError
        
    
class Directory(dict, Item):
    
    def __init__(self, name, mountpoint="", items=None):
        Item.__init__(self, name, mountpoint)
        dict.__init__(self)
        self.__build_cache = items or dict()
        
    def __len__(self):
        return len(self.get_items())
        
    def add_item(self, path, type, *args):
        if not issubclass(type, Item):
            raise TypeError
        try:
            self[path] = type(path, self.mountpoint, *args)
        except:
            _logger.exception("Error while creating filesystem object")
        
    def get_items(self):
        return [".", ".."] + map(lambda x: drop_root(x, len(self.name)), \
            set(self.keys()) | set(self.__build_cache.keys()))
            
    def _test_items(self):
        print set(self.keys())
        print set(self.__build_cache.keys())
        print set(self.keys()) | set(self.__build_cache.keys())
        print map(lambda x: drop_root(x, len(self.name)), set(self.keys()) | set(self.__build_cache.keys()))
    
    def getattr(self, fh=None):
        now = int(time.time())
        return dict(
            st_mode = (stat.S_IFDIR | 0755),
            st_ctime = now,
            st_mtime = now,
            st_atime = now,
            st_nlink = len(self)
        )
        
    def __getitem__(self, path):
        try:
            return dict.__getitem__(self, path)
        except KeyError, error:
            try:
                type = self.__build_cache.pop(path)
            except KeyError:
                _logger.error("path: %s" %path)
                _logger.error("self.keys(): %r" %self.keys())
                _logger.error("cache.keys(): %r" %self.__build_cache.keys())
                raise error
            if isinstance(type, tuple):
                self.add_item(path, type[0], *type[1:])
            else:
                self.add_item(path, type)
            return self[path]
        
    def __repr__(self):
        return "<%s '%s' containing %r>" %(self.__class__.__name__, self.name, self.get_items())
        
        
class Journal(Directory):
    
    @staticmethod
    def get_min_max_day():
        """ returns a two day-tuples containing the first and last item """
        today = date.today().timetuple()[:3]
        min_event = get_dbus_interface().FindEvents(0, 0, 1, True, "item", [])
        if min_event:
            min_event = min_event.pop()
            min_day = date.fromtimestamp(float(min_event["timestamp"])).timetuple()[:3]
        else:
            min_day = (today[0] - 1, today[1], today[2])
        return (min_day, today)
    
    def __init__(self, name, mountpoint=""):
        min_day, max_day = Journal.get_min_max_day()
        _logger.debug("constructing filesystem in range %r - %r" %(min_day, max_day))
        super(Journal, self).__init__(
            name,
            mountpoint,
            dict(
                (os.path.join(name, str(i)), (Year, min_day, max_day)) \
                for i in range(min_day[0], max_day[0] + 1)
            )
        )
        
    def __getitem__(self, path):
        try:
            return super(Journal, self).__getitem__(path)
        except KeyError, error:
            try:
                year_path = os.path.sep.join(path.split(os.path.sep)[:3])
            except:
                raise error
            return super(Journal, self).__getitem__(year_path)[path]
        
         
class Year(Directory):
    
    def __init__(self, name, mountpoint="", min_day=None, max_day=None):
        if None in (min_day, max_day):
            timerange = range(1, 13)
        else:
            min_month, max_month = (1, 13)
            year = int(name.split(os.path.sep)[-1])
            if min_day[0] == year:
                min_month = min_day[1]
            if max_day[0] == year:
                max_month = max_day[1] + 1
            timerange = range(min_month, max_month)        
        super(Year, self).__init__(
            name,
            mountpoint,
            dict(
                (os.path.join(name, "%.2i" %i), (Month, min_day, max_day)) for i in timerange
            )
        )
        
    def __getitem__(self, path):
        try:
            return super(Year, self).__getitem__(path)
        except KeyError:
            month = path.split(os.path.sep)
            if len(month) == 4:
                m = ["%.2i" %int(month.pop())]
                month_path = os.path.sep.join(month + m)
                return super(Year, self).__getitem__(month_path)
            elif len(month) == 5:
                day = month.pop()
                m = ["%.2i" %int(month.pop())]
                month_path = os.path.sep.join(month + m)
                return super(Year, self).__getitem__(month_path)[path]
            else:
                root = month[:5]
                day = root.pop()
                m = ["%.2i" %int(root.pop())]
                month_path = os.path.sep.join(root + m)
                #~ content = os.path.sep.join(month[5:])
                _logger.debug("self.keys() in Year.__getitem__:else %r" %self.keys())
                return super(Year, self).__getitem__(month_path)[path]
    
class Month(Directory):
    
    @staticmethod
    def get_dayrange(year, month, min_day=1, max_day=None):
        days = calendar.mdays[month]
        if month == 2:
            days += int(calendar.isleap(year))
        if max_day is not None and max_day < days:
            days = max_day
        days += 1
        return range(min_day, days)
            
    def __init__(self, name, mountpoint="", min_day=None, max_day=None):
        x = name.split(os.path.sep)
        year, month = map(int, x[-2:])
        if None in (min_day, max_day):
            timerange = Month.get_dayrange(year, month)
        else:
            min_d = 1
            max_d = None
            if min_day[0] == year and min_day[1] == month:
                min_d = min_day[2]
            if max_day[0] == year and max_day[1] == month:
                max_d = max_day[2]
            timerange = Month.get_dayrange(year, month, min_d, max_d)
        super(Month, self).__init__(
            name,
            mountpoint,
            dict((os.path.join(name, "%.2i" %i), Day) for i in timerange)
        )
        
    def __getitem__(self, path):
        try:
            return super(Month, self).__getitem__(path)
        except KeyError:
            day = path.split(os.path.sep)
            if len(day) == 5:
                n_day = ["%.2i" %int(day.pop())]
                n_month = ["%.2i" %int(day.pop())]
                day_path = os.path.sep.join(day + n_month + n_day)
                return super(Month, self).__getitem__(day_path)
            else:
                root = day[:5]
                filename = day[5:]
                n_day = ["%.2i" %int(root.pop())]
                n_month = ["%.2i" %int(root.pop())]
                day_path = os.path.sep.join(root + n_month + n_day)
                path = os.path.sep.join(root + n_month + n_day + filename)
                return super(Month, self).__getitem__(day_path)[path]
                
    
class Day(Directory):
    
    ##defines the shown content
    # 'File': local files
    # 'Web': web history and bookmarks
    CONTENT = ["File", "Web"]
    
    @staticmethod
    def make_safe_filename(uri):
        uri = str(uri.encode("UTF-8", "replace")) #fuse needs string not unicode
        uri = uri.split("///", 1)[-1]
        #~ uri = uri.replace(" ", "\ ")
        uri = uri.replace("/", "_")
        return uri
        
    @staticmethod
    def make_desktop_filename(uri):
        return "%s.desktop" %Day.make_safe_filename(uri)
    
    def __init__(self, name, mountpoint=""):
        super(Day, self).__init__(name, mountpoint)
        self.__got_events = False
        self.__date_tuple = map(int, name.split(os.path.sep)[-3:])
        self.add_item(
            os.path.join(self.name, get_random_name(suffix=".desktop")),
            VisitedPages,
            self.__date_tuple
        )
        self.__visited_index = None
        self.__visited_pages = None
        
    def getattr(self, *args):
        if not self.__got_events:
            try:
                self._generate_index()
            finally:
                self.__got_events = True
        return super(Day, self).getattr(*args)
        
    def __getitem__(self, path):
        if os.path.basename(path) == "visited_pages.html":
            if self.__visited_index is None:
                self._generate_index()
            return self.__visited_index
        else:
            return super(Day, self).__getitem__(path)
            
    def _generate_index(self):
        if self.__visited_pages is None:
            self._get_events()
        self.__visited_index = VisitedPagesIndex.generate_index(
            os.path.join(self.name, "visited_pages.html"),
            self.mountpoint,
            self.__visited_pages,
            *self.__date_tuple
        )
        
    def _get_events(self):
        year, month, day = map(int, self.name.split(os.path.sep)[-3:])
        current_day = date(year, month, day)
        self.__min_timestamp = int(time.mktime(current_day.timetuple()))
        self.__max_timestamp = self.__min_timestamp + 24 * 60 * 60
        events = get_dbus_interface().FindEvents(
            self.__min_timestamp,
            self.__max_timestamp,
            0,
            True,
            "item",
            [{"content": self.CONTENT}],
        )
        applications = set()
        annotations = dict()
        self.__visited_pages = _urldict()
        for event in events:
            uri = event["uri"]
            content = event["content"]
            if uri.startswith("file://%s" %self.mountpoint):
                continue
            if content == "File":
                filepath = os.path.join(self.name, Day.make_safe_filename(event["uri"]))
                self.add_item(
                    filepath,
                    EventFileProxy,
                    event
                )
                annotations[os.path.basename(filepath)] = {
                    "tags": event["tags"].split(", "),
                    "bookmarked": bool(event["bookmark"]),
                }
            elif content == "Web":
                if event["text"] in event["uri"]:
                    ## try to ignore cryptic captions
                    # this is why we skip title which are part of the url itself
                    # this is the case for some pages where tokens are
                    # shown in the title
                    continue
                splitresult = urlparse.urlsplit(event["uri"])
                self.__visited_pages.get_or_list(splitresult.netloc).append(event)
            else:
                raise TypeError("cannot handle event with content type == %r" %content)
            applications.add(str(event["app"]))
        for app in applications:
            self.add_item(
                os.path.join(self.name, os.path.basename(app)),
                AppProxy,
                app
            )
        self.add_item(
            os.path.join(self.name, ".annotations"),
            Annotations,
            annotations
        )
    
    
class File(Item):
    
    def getattr(self, fh=None):
        now = int(time.time())
        return dict(
            st_mode = (stat.S_IFREG | 0722),
            st_nlink = 1,
            st_size = 0,
            st_ctime = now,
            st_mtime = now,
            st_atime = now,
        )
        
    def __repr__(self):
        return "<File '%s'>" %self.name
        
    def read(self, length, offset):
        raise NotImplementedError
        
    def truncate(self, size=0):
        raise NotImplementedError
        
    def flush(self):
        pass
        
    def release(self):
        pass
        
        
class ContentFile(File):
    
    def __init__(self, name, mountpoint="", content=None):
        if content is None:
            raise ValueError("ContentFile needs content")
        super(ContentFile, self).__init__(name, mountpoint)
        self.__content = content.encode("UTF-8", "replace")
        self.__saved_content = self.__content
        now = int(time.time())
        self.__ctime = now
        self.__mtime = now
        self.__atime = now
        
    @property
    def _saved_content(self):
        return self.__saved_content
        
    def getattr(self, fh=None):
        result = super(ContentFile, self).getattr(fh)
        result["st_mode"] = (stat.S_IFREG | 0444)
        result["st_size"] = len(self)
        result["st_uid"] = os.getuid()
        result["st_gid"] = os.getgid()
        # adjust times
        result["st_ctime"] = self.__ctime
        result["st_mtime"] = self.__mtime
        result["st_atime"] = self.__atime
        return result
        
    def read(self, length, offset):
        self.__atime = int(time.time())
        return self.__content[offset:offset+length]
        
    def truncate(self, size=0):
        self.__saved_content = self.__content
        self.__content = self.__content[:size]
        self.__mtime = int(time.time())
        
    def write(self, data, offset):
        self.truncate(offset)
        self.__content += data
        self.__saved_content = self.__content
        self.__mtime = int(time.time())
        return len(self)
        
    def __len__(self):
        return len(self.__content)
        
        
class Annotations(ContentFile):
    
    def __init__(self, name, mountpoint="", annotations=None):
        self.__annotations = annotations
        if self.__annotations is None:
            self.__annotations = dict()
        content = json.dumps(self.__annotations, sort_keys=True, indent=4)
        super(Annotations, self).__init__(name, mountpoint, content)
        
    def getattr(self, fh=None):        
        result = super(Annotations, self).getattr(fh)
        result["st_mode"] = (stat.S_IFREG | 0644)
        return result
        
    def write(self, data, offset):
        if offset != 0:
            raise ValueError("Cannot handle huge files yet, offset=%i" %offset)
        try:
            data = self._validate(data)
        except ValueError:
            # revert to backup
            self.write(self._saved_content, offset)
            raise
        return super(Annotations, self).write(data, offset)
        
    def _validate(self, data):
        new_annotations = json.loads(data)
        if set(new_annotations.keys()) != set(self.__annotations.keys()):
            raise ValueError("You are not allowed to change the annotated files")
        # bookmarked and tags are the only allowed key for the sub-dict
        for annotations in new_annotations.values():
            if set(annotations.keys()) != set(["bookmarked", "tags"]):
                raise ValueError("Each file needs a 'bookmarked' and 'tags' annotation")
        for filename in new_annotations:
            # check bookmarked
            bookmarked = new_annotations[filename]["bookmarked"]
            if not isinstance(bookmarked, bool):
                raise ValueError("'bookmarked' annotation of '%s' needs to be a boolean value" %filename)
            if bookmarked != self.__annotations[filename]["bookmarked"]:
                self._change_annotation("change bookmark", filename, bookmarked) #TBD
            # check tags
            tags = new_annotations[filename]["tags"]
            if not isinstance(tags, list):
                raise ValueError("'tags' annotation of '%s' needs to be a list of strings" %filename)
            removed = set(self.__annotations[filename]["tags"]) - set(tags)
            if removed:
                self._change_annotation("remove tags", filename, removed) #TBD
            added = set(tags) - set(self.__annotations[filename]["tags"])
            if added:
                self._change_annotation("added tags", filename, added) #TBD
        return data
                
    def _change_annotation(self, *args):
        # TBD, waiting for API methods in zeitgeist
        print args
        
        
class VisitedPagesIndex(ContentFile):
    
    HEADER = textwrap.dedent("""\
    <html>

    <head>

        <title>Visited Pages for %(year)i/%(month)i/%(day)i</title>

        <style type="text/css">
            body { font:10pt Verdana; }
            a { color:blue; }
            #content { background-color:#dddddd; width:200px; margin-top:2px; }
        </style>

    </head>

    <body>

    <script type="text/javascript">

        function toggle(id) {
          var e = document.getElementById(id);

          if (e.style.display == '')
            e.style.display = 'none';
          else
            e.style.display = '';
        }

        function toggle2(id, link) {
          var e = document.getElementById(id);

          if (e.style.display == '') {
            e.style.display = 'none';
            link.innerHTML = 'Expand';
          } else {
            e.style.display = '';
            link.innerHTML = 'Collapse';
          }
        }

    </script>
    
    <h3>Visited Pages for %(year)i/%(month)i/%(day)i</h3></br>
    <i>(click on the domains to expand them)</i>
    
    """)
    
    @classmethod
    def generate_index(cls, name, mountpoint, visited_pages, year, month, day):
        content = cls.HEADER %{"year": year, "month": month, "day": day}
        content += "\t<ul>\n"
        for i, (netloc, pages) in enumerate(visited_pages.sorted_iter()):
            content += "\t\t<li><a href='javascript:;' onclick=\"toggle('node%i')\">%s</a></li>\n" %(i, netloc)
            content += "\t\t\t<ul id='node%i' style='display:none'>\n" %i
            for page in pages:
                content += "\t\t\t\t<li><a href='%s'>%s</a></li>\n" %(page["uri"], page["text"])
            content += "\t\t\t</ul>\n"
        content += "\t</ul>"
        content += "</body>\n</html>"
        return cls(name, mountpoint, content)        
    

class Starter(File):
    pass
        
        
class FileProxy(File):
    
    def __init__(self, name, mountpoint="", path=None):
        super(FileProxy, self).__init__(self, mountpoint)
        if not path:
            raise ValueError("FileProxy needs a target on the 'real' filesystem")
        if not os.path.isfile(path):
            raise OSError("Cannot create FileProxy object, %r is not a file" %path)
        self.__path = path
        ### THIS deadlocks if self.__path points to the vfs itself! ####
        if self.__path.startswith(mountpoint):
            raise RuntimeError("Can't show events for the vfs itself")
        self.__stat = os.stat(self.__path)
        
    def getattr(self, fh=None):
        return dict(
            st_mode = self.__stat.st_mode,
            st_nlink = self.__stat.st_nlink,
            st_size = self.__stat.st_size,
            st_ctime = self.__stat.st_ctime, #event time here?
            st_mtime = self.__stat.st_mtime, #event time here?
            st_atime = self.__stat.st_atime, #event time here?
            st_uid = self.__stat.st_uid,
            st_gid = self.__stat.st_gid,
        )
        
        
    def open(self, flags):
        self.__fd = os.open(self.__path, flags)
        return self.__fd
        
    def read(self, length, offset):
        return os.read(self.__fd, length)
        
    def release(self):
        os.close(self.__fd)
        

class EventFileProxy(FileProxy):
    
    def __init__(self, name, mountpoint="", event=None):
        if event is None:
            raise ValueError("EventFileProxy needs an event argument")
        path = "/%s" %event["uri"].split("///")[-1]
        super(EventFileProxy, self).__init__(name, mountpoint, path)


class AppProxy(FileProxy):
    
    def __init__(self, name, mountpoint="", application=None):
        if not application:
            raise ValueError("AppProxy needs an application argument")
        super(AppProxy, self).__init__(name, mountpoint, application)
    
    
class Symlink(File):
    
    def getattr(self, fh=None):
        attr = super(Symlink, self).getattr(fh)
        attr["st_mode"] = (stat.S_IFLNK | 0722)
        return attr
        
    def __repr__(self):
        return "<SymLink '%s'>" %self.name
        
    @property
    def target(self):
        raise NotImplementedError
        
        
class URLLink(Starter):
    
    def __init__(self, name, mountpoint, event):
        super(URLLink, self).__init__(name, mountpoint)
        self.__event = event
        self.__content = WWWLink(event[1], event[2]).write().getvalue()
        
    def getattr(self, fh=None):
        result = super(URLLink, self).getattr(fh)
        result["st_mode"] = (stat.S_IFREG | 0755)
        result["st_size"] = len(self.__content)
        return result
        
    def read(self, length, offset):
        return self.__content[offset:offset+length]
    
        
class VisitedPages(Starter):
    
    def __init__(self, name, mountpoint, date):
        super(VisitedPages, self).__init__(name, mountpoint)
        self.__content = WWWLink(
            "file://" + os.path.join(
                self.mountpoint,
                os.path.dirname(name).lstrip("/"),
                "visited_pages.html"
            ),
            "Visited URLs for %i/%i/%i" %tuple(date)
        ).write().getvalue()
        
    def getattr(self, fh=None):
        result = super(VisitedPages, self).getattr(fh)
        result["st_mode"] = (stat.S_IFREG | 0755)
        result["st_size"] = len(self.__content)
        result["st_uid"] = os.getuid()
        result["st_gid"] = os.getgid()
        return result
        
    def read(self, length, offset):
        return self.__content[offset:offset+length]
        
        
class Today(Starter):
    
    def __init__(self, name, mountpoint):
        super(Today, self).__init__(name, mountpoint)
        today = date.today()
        target = os.path.join(
            mountpoint, JOURNAL_IDENTIFIER, str(today.year), "%.2i" %today.month, "%.2i" %today.day
        )
        self.__content = StarterLink(target, TODAY_NAME).write().getvalue()
        
    def getattr(self, fh=None):
        result = super(Today, self).getattr(fh)
        result["st_mode"] = (stat.S_IFREG | 0755)
        result["st_size"] = len(self.__content)
        return result
        
    def read(self, length, offset):
        return self.__content[offset:offset+length]
        
        
class TodaySymlink(Symlink):
    
    @property
    def target(self):
        today = date.today()
        return os.path.join(ROOT_JOURNAL, str(today.year), "%.2i" %today.month, "%.2i" %today.day)
    
    
class Root(Directory):
    """ The root of the zeitgeist-filesystem
    It contains a /journal/ directory which structures files in a
    calendar like way and a `today` link, which links to the journal
    entry for the current day.
    If desktop is True the `today` link is a .desktop file called
    /today.desktop otherwise /today is a symlink
    """
    
    def __init__(self, mountpoint="", desktop=False):
        root = {ROOT_JOURNAL: Journal}
        if desktop:
            root["%s.desktop" %ROOT_TODAY] = Today
        else:
            root[ROOT_TODAY] = TodaySymlink
        super(Root, self).__init__(
            os.path.sep,
            mountpoint,
            root
        )

    def __getitem__(self, path):
        try:
            if path == os.path.sep:
                return self
            elif path == ROOT_JOURNAL or (path.startswith(ROOT_TODAY) and path.count("/") == 1):
                return super(Root, self).__getitem__(path)
            elif path.startswith("/.Trash"):
                return Directory(path)
            elif "/." in path and not path.endswith(".annotations"):
                #TBD do we need this, how can we handle/use .hidden files
                # but we need to show .annotations!
                raise OSError(errno.ENOENT, "")
            else:
                return super(Root, self).__getitem__(ROOT_JOURNAL)[path]
        except KeyError:
            _logger.exception("KeyError in Root.__getitem__")
            raise OSError(errno.ENOENT, "")
