# -*- coding: utf-8 -*-
#
# (C) Copyright 2009, Markus Korn <thekorn@gmx.de>
#
# ##################################################################
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ##################################################################

import os
import json

from zeitgeistfs import MOUNT_FILE

def get_mountpoint():
    if not os.path.exists(MOUNT_FILE):
        return None
    else:
        with open(MOUNT_FILE) as f:
            return f.read()
            
MOUNTPOINT = get_mountpoint()

def is_in_zeitgeistfs(path):
    if not os.path.isfile(path):
        # only files can be annotated
        return False
    elif MOUNTPOINT is None:
        # zeitgeistfs is not mounted
        return False
    elif not path.startswith(MOUNTPOINT):
        # file is not in zeitgeist filesystem
        return False
    elif path in ("/", "/journal", "/today") or os.path.basename(path) in (".annotations",): #TBD
        # cannot annotated reserved files
        return False
    elif not os.path.isfile(os.path.join(os.path.dirname(path), ".annotations")):
        # no .annotations file for this file found
        return False
    return True
    
def has_annotations(path):
    if not os.path.isdir(path):
        # only dirs can contain annotations
        return False
    elif MOUNTPOINT is None:
        # zeitgeistfs is not mounted
        return False
    elif not path.startswith(MOUNTPOINT):
        # dir is not in zeitgeist filesystem
        return False
    elif not os.path.isfile(os.path.join(path, ".annotations")):
        # no .annotations file for this file found
        return False
    return True
    
def read_annotations_for_dir(path):
    if not has_annotations(path):
        raise TypeError("%r has not zeitgeist annotations")
    annotations_file = os.path.join(path, ".annotations")
    try:
        with open(annotations_file) as f:
            annotations = json.load(f)
    except ValueError:
        raise RuntimeError("Unable to read .annotations for %r" %path)
    return annotations

def add_tags(path, new_tags):
    path = os.path.abspath(os.path.expanduser(path))
    if not is_in_zeitgeistfs(path):
        raise RuntimeError("Can't add tags to %r, it's not a file in the zeitgeist filesystem" %path)
    annotations_file = os.path.join(os.path.dirname(path), ".annotations")
    try:
        with open(annotations_file) as f:
            annotations = json.load(f)
    except ValueError:
        raise RuntimeError("Unable to read .annotations for %r" %path)
    try:
        tags = annotations[os.path.basename(path)]["tags"]
    except KeyError, e:
        raise RuntimeError("cannot find tags for %r" %path)
    for tag in new_tags:
        if tag not in tags:
            tags.append(tag)
    with open(annotations_file, "w") as f:
        json.dump(annotations, f, sort_keys=True, indent=4)
    
    
def remove_tags(path, tags):
    pass
